#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import h5py
import sys
import multiprocessing as mp
from tomophantom import TomoP3D
from tomophantom.supp.astraOP import AstraTools
from tomophantom.supp.artifacts import ArtifactsClass

g_filename=''
g_Io=0
g_dataset_name=''
maxim=0

def progress(count, total, status=''):
    bar_len = 80
    filled_len = int(round(bar_len * count / float(total)))

    percents = round(100.0 * count / float(total), 1)
    bar = '=' * filled_len + '-' * (bar_len - filled_len)

    sys.stdout.write('[%s] %s%s ...%s\r' % (bar, percents, '%', status))
    sys.stdout.flush()
    
def fun(itr):
	global g_filename
	global g_dataset_name
	global g_Io
	global maxim
	f2=h5py.File(g_filename,'r')
	sino = f2[g_dataset_name][:,itr,:]
	sino=sino/maxim
	sino=g_Io*np.exp(-sino)
	sino=np.random.poisson(sino)
	sino=sino/g_Io
	f2.close()
	return sino
    
'''
N_size:          number of horizontal detectors
P:               number of sinograms/vertical detectors
angles_num:      number of projectios/height of the sinograms
model:           model of the pin and salt-water (number 14 in the library)
model2:          model of the pin in order to remove the lower part of the salt-water ellipsoid (number 15 in the library)
model3:          model of the bubbles of hydrogen in the salt-water (number 17 and 19 for the training and testing dataset respectively)
model4:          model of the magnesium deposits in the pin (number 16 and 18 for the training and testing dataset respectively)
lib_path:        lib_path to the library (default './Phantom3DLibrary.dat')
filename:        name for the h5 file (default 'Synthetic_Pin_For_Training.h5')
dataset_name:    name for the dataset in the h5 file (default '/1-DarkFlatFieldCorrection-tomo/data')
number_workers:  number of parallel processes that will handle the addition of noise
Io:              level of Poisson Noise added
'''
    
def create_synthetic_projections(N_size = 1280, P = 1080, angles_num = 3601, model1 = 14, model2 = 15, model3 = 17, model4 = 16,
                                 lib_path = './Phantom3DLibrary.dat', filename = 'Synthetic_Pin_For_Training.h5',
                                 dataset_name = '/1-DarkFlatFieldCorrection-tomo/data', number_workers = 8, Io=25000):
	print ("Building 3D phantom using TomoPhantom software")
	global g_filename
	g_filename=filename
	global g_dataset_name
	g_dataset_name=dataset_name
	global g_Io
	g_Io=Io
	#This will generate a N_size x N_size x N_size phantom (3D)
	phantom_tm = TomoP3D.Model(model1, N_size, lib_path)
	phantom_tm2 = TomoP3D.Model(model2, N_size, lib_path)
	phantom_tm[phantom_tm2==1]=0.3
	phantom_tm2 = TomoP3D.Model(model3, N_size, lib_path)
	phantom_tm[phantom_tm2==1]=0
	phantom_tm2 = TomoP3D.Model(model4, N_size, lib_path)
	phantom_tm[phantom_tm2==1]=0.2
	phantom_tm=np.roll(phantom_tm,-80,axis=2)
	phantom_tm=np.roll(phantom_tm,-80,axis=1)

	# generate projection data
	angles = np.linspace(0,180,angles_num,dtype='float32')
	angles_rad = angles*(np.pi/180)

	Atools = AstraTools(N_size, angles_rad, N_size, 'gpu') # initiate a class object (gpu or cpu)
	proj_data_3D = np.empty((angles_num,P,N_size),'float32')
	f=h5py.File(filename,'w')
	dset = f.create_dataset(dataset_name, (angles_num,P,N_size), chunks=True)
	
	c=0
	global maxim
	for i in range(N_size-P,N_size): #The phantom created by Tomophantom is a cube, however I create the N_size-P:N_size sinograms as the top ones don't contain features ("air")
		sino = Atools.forwproj(phantom_tm[i,:,:]) # generate numerical sinogram
		mx=sino.max()
		if mx>maxim:
			maxim=mx
		proj_data_3D[:,c,:]=sino
		c+=1
		progress(i, N_size, status=('Creating Projections'))
		
	dset[:,:,:]=proj_data_3D
	del proj_data_3D
	f.close()

	pool = mp.Pool(processes=number_workers)
	dset=np.stack(pool.map(fun, range(0,P)),axis=1)
	f3=h5py.File(filename,'a')
	f3[dataset_name][:,:,:]=dset
	f3.close()

if __name__=='__main__':

	create_synthetic_projections(1280, 1080, 3601, 14, 15, 17, 16, './Phantom3DLibrary.dat', 'Synthetic_Pin_For_Training.h5', '/1-DarkFlatFieldCorrection-tomo/data', 8, 25000)
